
from pretty_midi.pretty_midi import Note

from abc import abstractmethod


def ct_time_to_beat(time: float, tempo: int) -> int:
    b4 = 60 / tempo
    b8 = b4 / 2
    b16 = b8 / 2
    b32 = b16 / 2

    beat, sub = calc_time_to_beat(time, b32)

    return beat


def calc_time_to_beat(time, beat_time) -> (int, int):
    main_beat: int = time // beat_time
    sub_time: int = time % beat_time
    return main_beat, sub_time


class Token:
    def __init__(self, tempo: int, token_type: str):
        self.tempo = tempo
        self.token_type = token_type
        self.token_position = 0

    @abstractmethod
    def get_token(self, back_notes: Note, note: Note) -> int:
        pass

    @abstractmethod
    def get_range(self) -> int:
        pass

    @abstractmethod
    def convert(self) -> int:
        pass

    def __call__(self, back_notes: Note, note: Note, *args, **kwargs):
        symbol: int = self.get_token(back_notes, note)
        if symbol == -1:
            my_token = None
            pass
        else:
            my_token = f"{self.token_type}_{symbol}"

        return my_token


class Pitch(Token):

    def get_range(self) -> int:
        return 128

    def get_token(self, back_notes: Note, note: Note) ->int:
        p: int = note.pitch
        return p


class Velocity(Token):

    def get_token(self, back_notes: Note, note: Note) -> int:
        v: int = note.velocity
        return v

    def get_range(self) -> int:
        return 128


class Duration(Token):

    def get_token(self, back_notes: Note, note: Note) -> int:
        start = ct_time_to_beat(note.start, self.tempo)
        end = ct_time_to_beat(note.end, self.tempo)
        d = int(max(abs(end - start), 1))

        if 100 < d:
            d = 100

        return d

    def get_range(self) -> int:
        return 100


class Start(Token):
    def get_token(self, back_notes: Note, note: Note) -> int:
        s = ct_time_to_beat(note.start, self.tempo)
        return s % 32

    def get_range(self) -> int:
        return 32


class Shift(Token):

    def get_token(self, back_notes: Note, note: Note) -> int:
        if back_notes is None:
            return -1
        else:
            back_start = ct_time_to_beat(back_notes.start, self.tempo)
            note_start = ct_time_to_beat(note.start, self.tempo)

            shift = int(abs((back_start // 32) - (note_start // 32)))

            if shift > 3:
                shift = 3

        return shift

    def get_range(self) -> int:
        return 4


from abc import abstractmethod
import torch



class LearningProgress:

    @abstractmethod
    def step_optimizer(self, optimizer, model, accumulation_steps, **kwargs):
        pass

    @abstractmethod
    def get_device(self):
        pass


class _DefaultLearningProgress(LearningProgress):

    def get_device(self):
        if torch.cuda.is_available():
            return torch.device('cuda')
        else:
            return torch.device('cpu')
        pass

    def step_optimizer(self, optimizer, model, accumulation_steps, **kwargs):

        # パラメータの勾配を累積ステップ数でスケーリング
        for param in model.parameters():
            if param.grad is not None:
                param.grad.data /= accumulation_steps

        optimizer.step()  # オプティマイザを更新

        print(f"現在のNORMは{self.get_gradient_norm(model)}です。")

        optimizer.zero_grad()
        pass

    def get_gradient_norm(self, model):
        total_norm = 0
        # モデルのすべてのパラメータについて、勾配がNoneでないものを対象にする
        for param in model.parameters():
            if param.grad is not None:
                param_norm = param.grad.detach().data.norm(2)  # L2ノルムを計算
                total_norm += param_norm.item() ** 2  # 勾配ノルムの2乗を足す
        total_norm = total_norm ** 0.5  # 最終的に平方根をとってL2ノルムを計算
        return total_norm



import json
from typing import List
from numpy import ndarray

from . import constants
from .aya_node import Token, Start,Shift,Pitch,Velocity,Duration

'''旋律トークン'''
PITCH_TYPE = 'p'
VELOCITY_TYPE = 'v'
DURATION_TYPE = 'd'

'''指示トークン'''
START_TYPE = 's'
SHIFT_TYPE = 'h'


def get_token_converter(tempo: int) -> List[Token]:
    register: List[Token] = list()

    register.append(Shift(tempo, SHIFT_TYPE))
    register.append(Start(tempo, START_TYPE))
    register.append(Pitch(tempo, PITCH_TYPE))
    #register.append(Velocity(tempo, VELOCITY_TYPE))
    register.append(Duration(tempo, DURATION_TYPE))

    return register


class Tokenizer:
    def __init__(self, token: List[Token], load_data: str = None):
        if load_data is None:
            # 特殊トークン
            self.special_token_position = 3
            p = self.special_token_position
            for t in token:
                t.token_position = p
                p += t.get_range()
            self.token_list = token
            self.tokens: dict = dict()
            self.token_max: dict = self._init_mx_dict(p)
            self.tokens[constants.PADDING_TOKEN] = 0
            self.tokens[constants.START_SEQ_TOKEN] = 1
            self.tokens[constants.END_SEQ_TOKEN] = 2
            self.is_converter = False
        else:
            self.is_converter = True
            with open(load_data, 'r') as file:
                self.tokens: dict = json.load(file)
                self.rev_tokens: dict = {v: k for k, v in self.tokens.items()}

    def _init_mx_dict(self, mx) -> dict:
        my_dict = dict()
        for i in range(0, mx):
            my_dict[i] = 0
        return my_dict

    def rev_get(self, a):
        return self.rev_tokens[a]

    def get(self, token: str):
        if token is not None:
            if token not in self.tokens and "<" not in token:
                sp = token.split('_')
                length = self.get_length(sp[0])

                self.tokens[token] = length
            if not self.is_converter:
                self.token_max[self.tokens[token]] += 1
            return self.tokens[token]

    def get_length(self, token_type: str):
        for t in self.token_list:
            if t.token_type == token_type:
                p = t.token_position
                t.token_position += 1
                return p
        pass

    def save(self, save_directory):
        json_string = json.dumps(self.tokens)
        with open(save_directory + "/vocab_list.json", 'w') as file:
            file.write(json_string)

        json_s = json.dumps(self.token_max)
        with open(save_directory + "/vocab_max.json", 'w') as file:
            file.write(json_s)

    pass
